#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/uaccess.h>
#include <linux/unistd.h> 
#include <../arch/x86/include/asm/cacheflush.h>

#include "../find.c" 
#include "../CR0.c" 
#include "syscall.h" 

asmlinkage long new_sys_call( const char __user *buf, size_t count ) {
   asm( "movl $0, %%eax\n"  // эквивалент return 0;
        "popl %%ebp    \n"
        "ret           \n"
        "L2: nop       \n"  // нам нужна метка L2 после return
        ::: "%eax"
      );
   return 0;  // только для синтаксиса
};

static int __init new_sys_init( void ) {
   void *addr_sys_call; 
   size_t sys_size;
   long end; 
   void **taddr;                   // адрес таблицы sys_call_table 
   if( ( taddr = find_sym( "sys_call_table" ) ) != NULL )
      printk( "! адрес sys_call_table = %p\n", taddr );
   else 
      return -EINVAL | printk( "! sys_call_table не найден\n" );
   asm( "movl $L2, %%eax \n"
        :"=a"(end)::  
      ); 
   printk( "! адрес статической функции %p\n", new_sys_call );
   printk( "! адрес конца кода функции %lx\n", end );
   sys_size = end - (long)new_sys_call;
   printk( "! размер функции обработчика %d байт\n", sys_size );
   // выделяем блок памяти под функцию обработчик
   addr_sys_call = kmalloc( sys_size, GFP_KERNEL );
   if( !addr_sys_call ) return -EINVAL | printk( "! memory allocation error!\n" );
   printk( "! выделен блок %d байт с адреса %p\n", sys_size, addr_sys_call );
   // копируем резидентный код нового обработчика в выделенный блок памяти
   memcpy( addr_sys_call, new_sys_call, sys_size );
   // снять бит NX-защиты с страницы
   //int set_memory_x(unsigned long addr, int numpages);
   set_memory_x( (long unsigned)addr_sys_call, 1 );
   printk( "! адрес нового sys_call = %p\n", addr_sys_call );
   rw_enable();
   taddr[ __NR_own ] = addr_sys_call;
   rw_disable();
   return -EPERM;
}
module_init( new_sys_init );

MODULE_LICENSE( "GPL" );
MODULE_AUTHOR( "Oleg Tsiliuric <olej@front.ru>" );
